import java.util.*;
import java.io.*;
import java.awt.Color;
import org.faceless.pdf2.*;

/**
 * A more complex example which creates a Unicode Character Map, showing
 * every character in the font. Not terribly efficient, but good as a
 * demonstration, and actually useful as well.
 */
public class CharacterMap
{
    // This method actually does the grunt work of creating the PDF page.
    // If you're reading this source to find out about creating PDF's, this
    // routine is the main, if not the only one you need to look at.
    // Consequently we've put it at the top of the class so you can't miss
    // it, at the expense of clarity of the overall source.
    //
    // start - the first character on the page
    // end - the last character on the page. No more than 128 after the start.
    // name - the name of the page (for the title).
    //
    private static PDFPage addPage(PDF p, char start, char end, String name, PDFFont f)
    {
	final int mod=16;	 // Number of characters per column.

	// Create the page
	//
	PDFPage page = p.newPage(PDF.PAGESIZE_A4);
   //	page.setFilter(PDFPage.FILTER_NONE);	// For debugging

	// Create some styles. For efficiency these should be created
	// once, elsewhere, and passed in. For clarity, they aren't.
	//
	// "fontstyle" is used to actually render the glyphs.
	//
	PDFStyle fontstyle = new PDFStyle();
	fontstyle.setFillColor(Color.black);
	fontstyle.setFont(f,12);
	fontstyle.setTextAlign(PDFStyle.TEXTALIGN_CENTER);

	// "titlestyle". Guess what this is for.
	//
	PDFStyle titlestyle = new PDFStyle();
	titlestyle.setFont(new StandardFont(StandardFont.HELVETICA), 18);
	titlestyle.setFillColor(Color.black);
	titlestyle.setTextAlign(PDFStyle.TEXTALIGN_CENTER);

	// "numberstyle" is for the "U+xxxx" notes in each character box
	//
	PDFStyle numberstyle = new PDFStyle();
	numberstyle.setFont(new StandardFont(StandardFont.HELVETICAOBLIQUE), 6);
	numberstyle.setFillColor(Color.black);
	numberstyle.setTextAlign(PDFStyle.TEXTALIGN_RIGHT);

	// "hollowboxstyle" is to surround glyphs that actually exist
	//
	PDFStyle hollowboxstyle = new PDFStyle();
	hollowboxstyle.setLineColor(Color.black);

	// "greyboxstyle" is to surround glyphs that don't exist.
	//
	PDFStyle greyboxstyle = (PDFStyle)hollowboxstyle.clone();
	greyboxstyle.setFillColor(Color.gray);


	// Create the page content.



	// Add the title, centered on the page 50 points from the top.
	//
	page.setStyle(titlestyle);
	page.drawText(name+" ("+toHex(start)+" - "+toHex(end)+")", page.getWidth()/2, page.getHeight()-50);

	// Crucially, the page is measured in PERCENT. This makes laying out grids
	// much easier.
	//

    page.setUnits(PDFPage.UNITS_PERCENT , PDFPage.ORIGIN_PAGETOP);

	float h = (95/mod);	     // height of each box, in percent
	float w = (95/8);	     // width of each box, in percent

    int left = 10;           // Position to start drawing table in percent
    int top =  10;           //


	for (char c=start; c<end; c++)
	{
	    // Calculate row and column for this glyph.
	    //
	    int row=(( c - start) % mod);
	    int col=(int)Math.floor(( c - start) / mod);

	    // If the character is defined, draw a hollow box, then
	    // place the glyph in the center.
	    //
	    // If it's not defined, draw a gray box
	    //
	    if (f.isDefined(c)) {
		page.setStyle(hollowboxstyle);
		page.drawRectangle(left+(col*w),top+(row*h),left + ((col+1)*w),top + ((row+1)*h));
		page.setStyle(fontstyle);
		page.drawText(""+c, left+((col+0.5f)*w), top+((row+0.6f)*h));
	    } else {
		page.setStyle(greyboxstyle);
		page.drawRectangle( left+(col*w), top+(row*h),left + ((col+1)*w),top +((row+1)*h));
	    }

	    // Add the Unicode number to the box at the bottom right
	    //
	    page.setStyle(numberstyle);
	    page.drawText(toHex(c), left+((col+0.9f)*w), top+((row+0.9f)*h));
	}
	return page;
    }



    //-----------------------------------------------------------------------
    // Rest of the program continues from here on.



    // The start of the Unicode code blocks. Lifted gratefully from
    // the source to java.lang.Character.UnicodeBlock.
    //
    private static final char[] blocks = { '\u0000', '\u0080', '\u0100', '\u0180', '\u0250', '\u02B0', '\u0300', '\u0370', '\u0400', '\u0500', '\u0530', '\u0590', '\u0600', '\u0700', '\u0900', '\u0980', '\u0A00', '\u0A80', '\u0B00', '\u0B80', '\u0C00', '\u0C80', '\u0D00', '\u0D80', '\u0E00', '\u0E80', '\u0F00', '\u0FC0', '\u10A0', '\u1100', '\u1200', '\u1E00', '\u1F00', '\u2000', '\u2070', '\u20A0', '\u20D0', '\u2100', '\u2150', '\u2190', '\u2200', '\u2300', '\u2400', '\u2440', '\u2460', '\u2500', '\u2580', '\u25A0', '\u2600', '\u2700', '\u27C0', '\u3000', '\u3040', '\u30A0', '\u3100', '\u3130', '\u3190', '\u3200', '\u3300', '\u3400', '\u4E00', '\uA000', '\uAC00', '\uD7A4', '\uD800', '\uE000', '\uF900', '\uFB00', '\uFB50', '\uFE00', '\uFE20', '\uFE30', '\uFE50', '\uFE70', '\uFF00' };


    public static void main(String[] args)
        throws IOException
    {
	// First load the font from the command line arguments.
	//
	PDFFont f = null;
	try {
	    f = selectFont(args);
	} catch (IOException e) {
	    e.printStackTrace();
	    System.exit(1);
	}

	// Create a new PDF
	//
	PDF p = new PDF();

	List bookmarks = p.getBookmarks();

	// Add some pages for each block to the document. There may
	// not be any characters defined for this block, in which case
	// addBlock() adds no pages and returns null.
	//
	for (int i=0;i<blocks.length-1;i++)
	{
	    Character.UnicodeBlock b = Character.UnicodeBlock.of(blocks[i]);
	    if (b!=null && !b.toString().equals("CJK_UNIFIED_IDEOGRAPHS") && !b.toString().equals("HANGUL_SYLLABLES")) {
		String blockname = b.toString();
		PDFPage first = addBlock(p, blocks[i], blocks[i+1], blockname, f);
		if (first!=null) {
		    bookmarks.add(new PDFBookmark(blockname, PDFAction.goTo(first)));
		}
	    }
	}

	// Set some meta-information. Google is indexing PDF's these
	// days, so this is more useful than you might think.
	//
	p.setInfo("Author", "Big Faceless Organization, Inc.");
	p.setInfo("Title", "Character map for \""+f.getBaseName()+"\"");
	p.setInfo("Subject", "Shows the Unicode codepoints for the \""+f.getBaseName()+"\" font.");
	p.setInfo("Keywords", f.getBaseName()+", Unicode, character map, font");

	// Write the document to a file
	//
	OutputStream fo = new FileOutputStream("CharacterMap.pdf");
	p.render(fo);
	fo.close();
    }


    //---------------------------------------------------------------------


    // Add some pages (if necessary) to the document for this Unicode Block.
    // If no characters are defined in this range, add nothing and return null,
    // otherwise return the first page created (for bookmarking)
    //
    // start - the first character in the blocks range
    // end - the last character in the blocks range
    // name - the name of the block (to be added as a title on each page)
    //
    private static PDFPage addBlock(PDF p, char start, char end, String name, PDFFont f)
    {
	PDFPage firstpage=null;
	boolean entries_in_this_block = false;

	// Establish if there are any entries in this range
	//
	if (Character.UnicodeBlock.of(start)!=null) {
	    for (char c=start; c<end; c++) {
	       if (f.isDefined(c)) {
		   entries_in_this_block = true;
	       }
	    }
	}

	// There are. Add one (or more) pages of 128 glyphs each.
	//
	if (entries_in_this_block)
	{
	    System.out.println("Rendering "+name);
	    for (char pagestart=start; pagestart<end; pagestart+=128)
	    {
		// Next, establish if there are characters on this page.
		// More efficient souls than myself would check this in
		// the loop above and store the results.
		//
		boolean entries_in_this_page = false;
		char pageend = (char)(pagestart+128);
		if (pageend>end) {
		    pageend = end;
		}
		for (char c=pagestart; c<pageend; c++) {
		    if (f.isDefined(c)) {
		        entries_in_this_page=true;
		    }
		}

		// Yes, there are entries on this page. Create it
		//
		if (entries_in_this_page) {
		    System.out.println("* "+toHex(pagestart)+" to "+toHex(pageend));
		    PDFPage page = addPage(p, pagestart, pageend, name, f);
		    if (firstpage==null) {
			firstpage=page;
		    }
		}
	    }
	}
	return firstpage;
    }


    // This unpleasant piece of code loads and returns the font
    // specified on the command line.
    //
    private static PDFFont selectFont(String[] args)
        throws IOException
    {
	PDFFont font = null;

	String ttfname=null, afmname=null, pfbname=null, stdname=null;
	String[] names = {  "Times-Roman",
			    "Times-Bold",
			    "Times-Italic",
			    "Times-BoldItalic",
			    "Helvetica",
			    "Helvetica-Bold",
			    "Helvetica-Oblique",
			    "Helvetica-BoldOblique",
			    "Courier",
			    "Courier-Bold",
			    "Courier-Oblique",
			    "Courier-BoldOblique",
			    "Symbol",
			    "ZapfDingbats"
			};

	for (int i=0;i<args.length;i+=2) {
	    if (args[i].equals("--ttf")) {
	        ttfname = args[i+1];
	    } else if (args[i].equals("--afm")) {
	        afmname = args[i+1];
	    } else if (args[i].equals("--pfb")) {
	        pfbname = args[i+1];
	    } else if (args[i].equals("--builtin")) {
	        stdname = args[i+1];
	    }
	}

	if (stdname!=null) {
	    for (int i=0;i<names.length;i++) {
	        if (names[i].equals(stdname)) {
		    font = new StandardFont(i);
		}
	    }
	    if (font==null) {
		throw new Error("No Such Standard Font: \""+stdname+"\"");
	    }
	} else if (ttfname!=null) {	// Assume TTF is to be embedded
	    font = new OpenTypeFont(new FileInputStream(ttfname), 2);
	} else if (afmname!=null) {
	    font = new Type1Font(new FileInputStream(afmname), (pfbname==null ? null : new FileInputStream(pfbname)));
	} else {
	    System.out.println("Usage: java CharacterMap [ --ttf <truetype font file> ] |");
	    System.out.println("                         [ --afm <AFM file> [ --pfb <PFB file> ]] |");
	    System.out.println("                         [ --builtin < Courier-Bold | Courier-BoldOblique |");
	    System.out.println("                                       Courier-Oblique | Courier |");
	    System.out.println("                                       Helvetica-Bold | Helvetica-BoldOblique |");
	    System.out.println("                                       Helvetica-Oblique | Helvetica | Symbol |");
	    System.out.println("                                       Times-Bold | Times-BoldItalic |");
	    System.out.println("                                       Times-Italic | Times-Roman |");
	    System.out.println("                                       ZapfDingbats > ]\n\n");
	    System.exit(0);
	}

	return font;
    }


    // Given a character, eg 'A', format it as "U+0041"
    //
    private static String toHex(char c)
    {
        String s = "0000" + Integer.toHexString((int)c).toUpperCase();
	return "U+"+s.substring(s.length()-4);
    }
}
