import org.faceless.pdf2.*;
import java.io.*;
import java.util.*;
import java.security.KeyStore;
import java.security.cert.X509Certificate;

// Usage:     java Dump <infile.pdf> [<password>]
//
// This example takes a PDF file on the commandline and
// dumps various bit of information, like the document
// meta information, lists of annotations, bookmarks, form
// elements and so on.
//
// $Id: Dump.java,v 1.8.2.1 2004/10/22 16:56:25 mike Exp $
//
public class Dump
{
    public static void main(String[] args) throws Exception
    {
        /**
         * If you have the DSE200 SDK, uncommenting this line will
         * allow you to verify documents timestamped with a DSE200.
         * See the DSE200Handler class for more information
         *
         * FormSignature.registerHandlerForVerification(new DSE200HandlerFactory(null,0));
         */

	if (args.length==0) {
	    System.err.println("USAGE: java Dump <infile.pdf> [<password>]\n");
	    System.exit(1);
	}

        PDFReader rd = new PDFReader(new FileInputStream(args[0]), (args.length>1 ? args[1] : ""));
	PDF p = new PDF(rd);

	System.out.println("DUMPING \""+args[0]+"\"\n----------------------------------------------------------------------\n");

	// Display the document information
	//
	System.out.println("Document ID: "+p.getDocumentID(true)+" "+p.getDocumentID(false));
	Map m = p.getInfo();
	System.out.println("Info dictionary:\n----------------------------");
	for (Iterator i = m.keySet().iterator();i.hasNext();) {
	    String k = (String)i.next();
	    if (!k.equals("_ModDate") && !k.equals("_CreationDate")) {
		System.out.println("  "+k+" = \""+m.get(k)+"\"");
	    }
	}

	for (int i=0;i<Event.ALL.length;i++) {
	    PDFAction act = p.getAction(Event.ALL[i]);
	    if (act!=null) System.out.println("Document "+Event.ALL[i]+"="+act);
	}

	// Display any "Named" actions defined in the document. These can
	// be referenced externally to open the document at a specific point.
	//
	m = p.getNamedActions();
	if (m.size()>0) {
	    System.out.println("\nNamed actions:\n----------------------------");
	    for (Iterator i = m.keySet().iterator();i.hasNext();) {
		String k = (String)i.next();
		System.out.println("  '"+k+"' = \""+m.get(k)+"\"");
	    }
	}

	// Display any XML Metadata defined in the document.
	//
	Reader r = p.getMetaData();
	if (r!=null) {
	    System.out.println("\nDocument XML Metainfo:\n---------------------");
	    BufferedReader br = new BufferedReader(r);
	    String s;
	    while ((s=br.readLine())!=null) {
	        System.out.println("  "+s);
	    }
	}

	// Display the documents bookmarks
	//
	List b = p.getBookmarks();
	if (b!=null) {
	    System.out.println("\nDocument Bookmarks:\n-----------------------");
	    dumpBookmarks(b, "");
	}

	// Display each page. For each page, show it's width, height, any open/close
	// actions, it's annotations and any XML meta data
	//
	System.out.println("\nDocument Pages:\n-----------------------");
	List l = p.getPages();
	for (int i=0;i<l.size();i++) {
	    System.out.println("\nPage "+(i+1)+"/"+l.size()+":\n-----------------");
	    PDFPage page = (PDFPage)l.get(i);
	    System.out.println("  Width="+page.getWidth()+"  Height="+page.getHeight());
	    System.out.println("  OpenAction="+page.getAction(Event.OPEN)+"  CloseAction="+page.getAction(Event.CLOSE));

	    // Display the page annotations
	    //
	    List annots = page.getAnnotations();
	    for (int j=0;j<annots.size();j++) {
		PDFAnnotation annot = (PDFAnnotation)annots.get(j);
                System.out.println("  Annotation "+(j+1)+"/"+annots.size()+": id="+annot.getUniqueID()+" "+annot);
	    }

	    // Display the page XML meta data, if any
	    //
	    r = page.getMetaData();
	    if (r!=null) {
		System.out.println("  XML Metainfo:\n---------------------");
		BufferedReader br = new BufferedReader(r);
		String s;
		while ((s=br.readLine())!=null) {
		    System.out.println("    "+s);
		}
	    }
	}

	// Print the contents of the documents form (if any). Very complete
	// but a little more complex than the rest of the example.
	//
	Map form = p.getForm().getElements();
	if (form.size()>0)
	{
	    System.out.println("\nDocument Form:\n-----------------------");
	    KeyStore keystore = FormSignature.loadDefaultKeyStore();

	    for (Iterator i=form.keySet().iterator(); i.hasNext();)
	    {
		String name=(String)i.next();
		FormElement el = p.getForm().getElement(name);
		List annots = el.getAnnotations();
		String val = el.getValue();

		// If it's a signature, verify the document.

		if (el instanceof FormSignature)
		{
		    final FormSignature sig = (FormSignature)el;
                    X509Certificate cert=null;
                    boolean verified=false, verifiable=false;
                    try
		    {
			verified = sig.verify();
			verifiable=true;
			if (sig.getSignatureHandler() instanceof PKCS7SignatureHandler)
			{
			    PKCS7SignatureHandler handler = (PKCS7SignatureHandler)sig.getSignatureHandler();
			    Calendar when = sig.getSignDate();
			    X509Certificate[] certs = handler.getCertificates();
			    cert = FormSignature.verifyCertificates(certs,keystore,null,when);
			} else {
			    System.out.println("Not a PKCS#7 public-key signature!");
			}
		    }
		    catch (Exception e) {
			System.out.println("ERROR: Verifying certificate: "+e);
		    }
		    String validity="";
		    if (sig.getNumberOfRevisionsCovered()!=p.getNumberOfRevisions()) {
		        validity = "Partial - covers the first "+sig.getNumberOfRevisionsCovered()+" of "+p.getNumberOfRevisions()+" revisions. Those covered are ";
		    }
		    if (verified && (sig.getSignatureHandler() instanceof PKCS7SignatureHandler && cert==null)) {
			validity += "Perfect (signature and certificates verified)";
		    } else if (verified) {
			validity += "Good (signature verified, but not certificates)";
		    } else if (verifiable) {
			validity += "Bad (signature not verified - document has been altered)";
		    } else {
			validity += "Unknown (unable to parse signature)";
		    }

		    System.out.println("  Signature: type="+sig.getFilter()+" signed="+sig.getSignDate().getTime());
		    System.out.println("             signed by \""+sig.getName()+"\" at \""+sig.getLocation()+"\" because \""+sig.getReason()+"\"");
		    System.out.println("             status: "+validity);
		    if (cert!=null) {
			String s = cert.toString();		// Indent certificate on screen
			for (int j=0;j<s.length();j++) {
			    if (s.charAt(j)=='\n') {
				s=s.substring(0,j+1)+"        "+s.substring(j+1);
				j+=8;
			    }
			}
			System.out.println("      Unverifiable certificate:\n        "+s);
		    }

		} else if (el instanceof FormText) {
		    String defaultval = ((FormText)el).getDefaultValue();
		    boolean password = (((FormText)el).getType() == FormText.TYPE_PASSWORD);
		    boolean multiline = (((FormText)el).getType() == FormText.TYPE_MULTILINE);

		    System.out.println("  Text");
		    System.out.print("  name=\""+name+"\"");
		    if (password) System.out.print(" password=\"true\"");
		    if (multiline) System.out.print(" multiline=\"true\"");
		    if (val!=null) System.out.print(" value=\""+replaceNL(val)+"\"");
		    if (defaultval!=null) System.out.print(" default=\""+replaceNL(defaultval)+"\"");
		    System.out.println();

		} else if (el instanceof FormButton) {
		    System.out.println("  Button");
		    System.out.print("  name=\""+name+"\"");
		    if (val!=null) System.out.print(" value=\""+val+"\"");
		    System.out.println();

		} else if (el instanceof FormCheckbox) {
		    System.out.println("  Checkbox");
		    System.out.print("  name=\""+name+"\"");
		    if (val!=null) System.out.print(" value=\""+val+"\"");
		    System.out.println();

		} else if (el instanceof FormRadioButton) {
		    System.out.println("  RadioButton");
		    System.out.print("  name=\""+name+"\"");
		    if (val!=null) System.out.print(" value=\""+val+"\"");
		    System.out.println();

		} else if (el instanceof FormChoice) {
		    Map opts = ((FormChoice)el).getOptions();
		    int type = ((FormChoice)el).getType();

		    System.out.println("  "+(type==FormChoice.TYPE_COMBO ? "Combo" : type==FormChoice.TYPE_SCROLLABLE ? "Scrollable" : "DropDown")+" Choice");
		    System.out.print("  name=\""+name+"\"");
		    if (val!=null) System.out.print(" value=\""+val+"\"");
		    System.out.println();

		    String out = "";
		    for (Iterator j=opts.keySet().iterator();j.hasNext();) {
			String optkey = (String)j.next();
			String optval = (String)opts.get(j);
			if (optval==null) {
			    out+=", \""+optkey+"\"";
			} else {
			    out+=", \""+optkey+"\"=\""+optval+"\"";
			}
		    }
		    System.out.println("  options=["+out.substring(2)+"]");
		}

		for (int j=0;j<Event.ALL.length;j++) {
		    PDFAction act = el.getAction(Event.ALL[j]);
		    if (act!=null) System.out.println("    "+Event.ALL[j]+"="+act);
		}
		for (int k=0;k<annots.size();k++) {
		    System.out.println("    "+formAnnotationString(el.getAnnotation(k)));
		}
		System.out.println();
	    }
	}
    }

    private static void dumpBookmarks(List l, String indent)
    {
        if (l==null) return;

	for (int i=0;i<l.size();i++) {
	    PDFBookmark b = (PDFBookmark)l.get(i);
	    // Fun with ASCII art follows
	    System.out.println(indent+"| ");
	    System.out.println(indent+(i==0?"+-":"+-")+b.getName()+": "+b.getAction());
	    dumpBookmarks(b.getBookmarks(), indent+(i==l.size()-1 ? "  ":"| "));
	}
    }

    // This method returns a String with information on the specified
    // Form Annotation. It's called from the Forms processing block, above
    //
    private static String formAnnotationString(WidgetAnnotation annot)
    {
	String out = "";

	float[] rect = annot.getRectangle();
	boolean visible = annot.isVisible();
	boolean printable = annot.isVisible();

	if (annot.getValue()!=null) {
	    out="\""+annot.getValue()+"\":  ";
	}
	out+="page="+annot.getPage().getPageNumber()+" rect="+rect[0]+","+rect[1]+"-"+rect[2]+","+rect[3];
	if (!visible) out+=" invisible";
	if (!printable) out+=" unprintable";
	out+="\n";

	// Print any events that are associated with the annotation.
	//
	for (int j=0;j<Event.ALL.length;j++) {
	    PDFAction act = annot.getAction(Event.ALL[j]);
	    if (act!=null) {
		out+="      "+Event.ALL[j]+"="+act+"\n";
	    }
	}
        return out.substring(0,out.length()-1);	// trim trailing '\n'
    }

    private static String replaceNL(String in)
    {
	String out = "";
        StringTokenizer st = new StringTokenizer(in, "\n", true);
	while (st.hasMoreTokens()) {
	    String z = st.nextToken();
	    if (z.equals("\n")) z="\\n";
	    out+=z;
	}
	return out;
    }
}
