// $Id: Sign.java,v 1.4 2004/05/16 21:49:52 mike Exp $

import java.util.*;
import java.io.*;
import org.faceless.pdf2.*;
import java.security.KeyStore;
import java.security.GeneralSecurityException;

/**
 * An example showing how to batch sign a PDF document read
 * from standard-in and written to standard-out.
 */
public class Sign
{
    private static String pdfpass, keystorename, reason, location, file;
    private static char[] storepass, keypass;

    private static String keystoretype="JKS", keystoreprovider=null, keyalias="mykey";
    private static SignatureHandlerFactory handler = FormSignature.HANDLER_SELFSIGN;

    public static void main(String[] args)
        throws IOException, GeneralSecurityException
    {
	// 1. Parse the arguments
	//
	getargs(args);
	if (keystorename==null) usage("No keystore specified");
	if (keypass==null && storepass!=null) keypass=storepass;

	// 2. Load the keystore. If a provider was specified, use it.
	//
	KeyStore keystore;
	if (keystoreprovider==null) {
	    keystore = KeyStore.getInstance(keystoretype);
	} else {
	    keystore = KeyStore.getInstance(keystoretype, keystoreprovider);
	}
	keystore.load(new FileInputStream(keystorename), storepass);

	// 3. Create the new signature
	//
	FormSignature sig = new FormSignature(keystore, keyalias, keypass, handler);
	sig.setReason(reason);
	sig.setLocation(location);
	PKCS7SignatureHandler pkcs7 =  (PKCS7SignatureHandler)sig.getSignatureHandler();
	sig.setName(FormSignature.getSubjectField(pkcs7.getCertificates()[0], "CN"));


	// 4. Add a visible appearance for this signature to the first page
	//

	// 4. Load the PDF from the specified file, apply the signature, give it
	// a visible appearance on the first page and write the PDF to "Sign.pdf"
	//
	InputStream in = new FileInputStream(file);
	PDF pdf = new PDF(new PDFReader(in, pdfpass));
	in.close();
	sig.addAnnotation(pdf.getPage(0), 100, 100, 300, 200);
	pdf.getForm().addElement("Test Signature", sig);
	OutputStream out = new FileOutputStream("Sign.pdf");
	pdf.render(out);
	out.close();
    }

    private static void getargs(String[] args)
    {
        for (int i=0;i<args.length;i++) {
	    if (args[i].startsWith("--pdfpassword=")) {
	        pdfpass = args[i].substring(14);
	    } else if (args[i].startsWith("--keypassword=")) {
	        keypass = args[i].substring(14).toCharArray();
	    } else if (args[i].startsWith("--storepassword=")) {
	        storepass = args[i].substring(16).toCharArray();
	    } else if (args[i].startsWith("--reason=")) {
	        reason = args[i].substring(9);
	    } else if (args[i].startsWith("--location=")) {
	        location = args[i].substring(11);
	    } else if (args[i].startsWith("--keyalias=")) {
	        keyalias = args[i].substring(11);
	    } else if (args[i].startsWith("--keystore=")) {
	        keystorename = args[i].substring(11);
	    } else if (args[i].startsWith("--keystoretype=")) {
	        keystoretype = args[i].substring(15);
	    } else if (args[i].startsWith("--keystoreprovider=")) {
	        keystoreprovider = args[i].substring(19);
	    } else if (args[i].equals("--handler=verisign")) {
	        handler = FormSignature.HANDLER_VERISIGN;
	    } else if (args[i].equals("--handler=selfsign")) {
	        handler = FormSignature.HANDLER_SELFSIGN;
	    } else if (args[i].equals("--handler=acrobat6")) {
	        handler = FormSignature.HANDLER_ACROBATSIX;
	    } else if (file==null) {
	        file = args[i];
	    } else {
	        usage("Unknown argument \""+args[i]+"\"");
	    }
	}
	if (file==null) usage("No filename specified");
    }

    private static void usage(String errorstring)
    {
	System.err.println("ERROR: "+errorstring);
	System.err.println("Usage: java Sign --keystore=<keystore file name>");
	System.err.println("                 --storepassword=<keystore password>");
	System.err.println("                 [--keyalias=<private key alias>]");
	System.err.println("                 [--keypassword=<private key password>]");
	System.err.println("                 [--pdfpassword=<PDF decryption password>]");
	System.err.println("                 [--keystoretype=<KeyStore type>]");
	System.err.println("                 [--keystoreprovider=<KeyStore provider>]");
	System.err.println("                 [--handler=verisign|selfsign]");
	System.err.println("                 [--location=<location comment>]");
	System.err.println("                 [--reason=<reason comment>] Input.pdf");
	System.err.println();
	System.err.println("--keystore          The filename of the KeyStore containing the private");
	System.err.println("                    key to sign the PDF document with");
	System.err.println("--storepassword     The password required to open the keystore.");
	System.err.println("--keyalias          (optional) The name of the private key in the keystore.");
	System.err.println("                    Defaults to \"mykey\"");
	System.err.println("--keypassword       (optional) The password used to access the private key.");
	System.err.println("                    defaults to the value of the \"storepassword\" option");
	System.err.println("--pdfpassword       (optional) The password required to decrypt the incoming");
	System.err.println("                    PDF, if required");
	System.err.println("--handler           (optional) The type of signature handler to use. Valid");
	System.err.println("                    values are \"selfsign\", to use the Adobe Self-Sign handler");
	System.err.println("                    or \"verisign\" to use the Verisign Document Signer.");
	System.err.println("                    The default is \"selfsign\"");
	System.err.println("--keystoretype      (optional) The type of KeyStore. Defaults to \"JKS\", for");
	System.err.println("                    \"Java KeyStore\". Other types like \"pkcs12\" may be");
	System.err.println("                    used provided an appropriate JCE provider is available");
	System.err.println("--keystoreprovider  (optional) The name of the provider supplying the KeyStore");
	System.err.println("                    implementation");
	System.err.println("--reason            (optional) The reason the document is being signed");
	System.err.println("--location          (optional) The location the document is being signed at");
	System.err.println();
	System.err.println("Signs a PDF document read from standard-in, and writes it to \"Sign.pdf\".");
	System.err.println("To create a self-signed JKS keystore called \"testkeystore\" suitable for use");
	System.err.println("with the Adobe self-sign handler, run the following command:");
	System.err.println();
	System.err.println("  keytool -genkey -keyalg RSA -sigalg MD5withRSA -keystore testkeystore");
	System.err.println();
	System.err.println("Be sure to specify a 2 letter code for the country, otherwise Acrobat will");
	System.err.println("be unable to validate the signature.");
	System.err.println();
	System.exit(0);
    }
}
