// $Id: DiskCache.java,v 1.7.6.2 2004/09/24 14:06:03 mike Exp $

package org.faceless.pdf2;

import java.io.*;

/**
 * A general purpose {@link Cache} which optionally writes closed streams
 * to files on disk. This class has been tested on UNIX and Windows under
 * multi-threaded environments, and we see no reason why it shouldn't work
 * under other platforms as well.
 * @see PDF#setCache
 * @since 2.2.2
 */
public class DiskCache implements Cache
{
    private static int next;            // The next number. Static just in case two DiskCaches
                                        // are created on the same directory.

    private String prefix;              // The directory to store the cached files in
    private int threshold;              // The smallest stream to write to disk

    /**
     * Create a new DiskCache
     *
     * @param prefix The prefix to begin the cached filenames with. For instance, if
     * prefix was "/tmp/cache/mycache.", the cache files would be "/tmp/cache/mycache.0",
     * "/tmp/cache/mycache.1" etc.
     *
     * @param threshold the minimum number of bytes that are considered worth caching to
     * disk, or 0 to prevent caching altogether
     */
    public DiskCache(String prefix, int threshold)
    {
        this.prefix = prefix;
        this.threshold = threshold==0 || prefix==null ? Integer.MAX_VALUE : threshold;
    }


    public Cache.Entry newEntry(int size)
    {
        return new DiskCacheEntry(size);
    }

    /**
     * Return the next filename to be used in the cache (of the form "prefixnnn")
     */
    private synchronized File nextFileName()
    {
        return prefix==null ? null : new File(prefix+(next++));
    }


    private class DiskCacheEntry extends OutputStream implements Cache.Entry
    {
        private OutputStream out;
        private File file;                      // null if data held in memory
        private int count;
        private boolean closed;

        public DiskCacheEntry(int size)
        {
            if (prefix!=null && size>threshold) {
                try {
                    toDisk();
                } catch (IOException e) {
                    throw new Error("Cannot create File \""+file+"\": "+e);
                }
            } else {
                out = new MyByteArrayOutputStream(size);
            }
        }

        public OutputStream getOutputStream()
        {
            if (closed) throw new IllegalStateException("Can't write to closed stream \""+file+"\"");
            return this;
        }

        public InputStream getInputStream()
        {
            if (file!=null) {
                try {
                    if (!closed) out.flush();
                    return new BufferedInputStream(new FileInputStream(file));
                } catch (IOException e) {
                    throw new Error("Cannot create FileInputStream for \""+file+"\": "+e);
                }
            } else {
                return new ByteArrayInputStream(((MyByteArrayOutputStream)out).buf, 0, count);
            }
        }

        public int size()
        {
            return count;
        }

        public void writeTo(OutputStream stream)
            throws IOException
        {
            if (out instanceof MyByteArrayOutputStream) {
                stream.write(((MyByteArrayOutputStream)out).buf,0,count);
            } else {
                InputStream in = getInputStream();
                byte[] b = new byte[8192];
                int l;
                while ((l = in.read(b))>=0) {
                    stream.write(b,0,l);
                }
                in.close();
            }
        }

        public Cache.Entry cloneEntry()
        {
            if (closed) {
                return this;
            } else {
                DiskCacheEntry copy = new DiskCacheEntry(count);
                try {
                    writeTo(copy.getOutputStream());
                } catch (IOException e) {
                    throw new Error("Cannot clone stream to \""+copy.file+"\": "+e);
                }
                return copy;
            }
        }

        public void close()
        {
            if (!closed) {
                try {
                    out.close();
                } catch (IOException e) {
                    throw new Error("Cannot close \""+file+"\": "+e);
                }
                closed=true;
            }
        }

        public String toString()
        {
            return "<filecache size="+size()+" file="+file+">";
        }

        public void finalize()
        {
            if (file!=null) file.delete();
        }

        //-----  OutputStream methods from here on ----

        public void write(int b) throws IOException
        {
            out.write(b);
            count++;
            if (file==null && count>threshold) toDisk();
        }

        public void write(byte[] b, int off, int len) throws IOException
        {
            if (file==null && count+len>threshold) toDisk();
            out.write(b,off,len);
            count+=len;
        }

        public void write(byte[] b) throws IOException
        {
            write(b,0,b.length);
        }

        private void toDisk()
            throws IOException
        {
            file = nextFileName();
            file.deleteOnExit();
            byte[] buf = out==null ? null : ((MyByteArrayOutputStream)out).buf;
            out = new BufferedOutputStream(new FileOutputStream(file));
            if (buf!=null) out.write(buf, 0, count);
        }
    }

    private final class MyByteArrayOutputStream extends OutputStream
    {
        private int count;
        byte[] buf;

        MyByteArrayOutputStream(int size)
        {
            buf = new byte[size==0 ? 16 : size];
        }

        public void write(int b)
        {
            if (count==buf.length) extend(count+8);
            buf[count++] = (byte)b;
        }

        public void write(byte[] b, int off, int len)
        {
            if (len==0) {
                return;
            } else if (off<0 || len<0 || off+len > b.length) {
                throw new IndexOutOfBoundsException(off+"+"+len+" outside "+b.length);
            }
            if (count+len > buf.length) extend(count+len);
            System.arraycopy(b, off, buf, count, len);
            count+=len;
        }

        public void write(byte[] b)
        {
            write(b,0,b.length);
        }

        /**
         * Increase the capacity of the internal buffer
         */
        private final void extend(int newlen)
        {
            if (newlen>buf.length) {
                if (newlen<buf.length+(buf.length>>1)) newlen=buf.length+(buf.length>>1);
                byte[] newbuf = new byte[newlen];
                System.arraycopy(buf, 0, newbuf, 0, count);
                buf=newbuf;
            }
        }
    }
}
